/**
 * TO DO
 *  - For some reason, the default arguments are not being translated into
 *	overloaded methods in Java.
 */

%module nds

/* Tell swig to include the header file in the C wrapper that it is creating,
 * but don't process the declarations in it yet.
 */
%{
#include "nds.h"
#include <stdio.h>
%}

/* Some built-in swig interface files that we will need. */
%include <exception.i>
%include <constraints.i>
#ifdef SWIGJAVA
%include "jniloader.i"
#endif

/* Special typedef to turn on special string handling for MATLAB.
 * MATLAB does not automatically convert java.lang.String return values to its
 * own char arrays, so for convenient calling from MATLAB we turn C strings into
 * double-scripted char arrays, which MATLAB knows to interpret as its own char
 * arrays. */
typedef char *c_str;
%{
	typedef char *c_str;
%}
%apply char * {c_str};

%typemap(javapackage) nds;

/* Make JNI classes package-private. */
%pragma(java) jniclassclassmodifiers = "class";

/* Strip package prefix from names */
%rename("%(strip:[nds_])s") "";

%ignore nds_connection;
%rename _nds_connection connection;

/* Strip NDS_ENUMNAME prefix from enum elements.  Inspired by exampple in
 * "SWIG Basics" documentation. */
%rename("%(regex:/NDS_DATA_TYPE_(.*)/\\1/)s", %$isenumitem) "";
%rename("%(regex:/NDS_CHANNEL_TYPE_(.*)/\\1/)s", %$isenumitem) "";
%rename("%(regex:/NDS_PROTOCOL_(.*)/\\1/)s", %$isenumitem) "";

/* Don't directly expose these. */
%ignore nds_errnum;
%ignore nds_open;
%ignore nds_close;
%ignore nds_get_host;
%ignore nds_get_port;
%ignore nds_get_protocol;
%ignore nds_find_channels;
%ignore nds_fetch;
%ignore nds_iterate;
%ignore nds_next;

#if defined(SWIGJAVA)
%include "nds_java.i"
#elif defined(SWIGPYTHON)
%include "nds_python.i"
#elif defined(SWIGOCTAVE)
%include "nds_octave.i"
#endif

/* Exception handling */
%{
#include <errno.h>
%}

/** Exception handling for routines that must return a non-NULL result. */
%define NDS_SWIG_EXCEPTION_NONNULL
	%exception %{
		$action
		if (!result)
		#if defined(SWIGJAVA)
			SWIG_JavaException(jenv, SWIG_RuntimeError, nds_strerror(errno));
		#else
			SWIG_exception(SWIG_RuntimeError, nds_strerror(errno));
		#endif
	%}
%enddef

/** Exception handling for routines that must return a non-negative result. */
%define NDS_SWIG_EXCEPTION_NONNEG
	%exception %{
		$action
		if (result < 0)
		#if defined(SWIGJAVA)
			SWIG_JavaException(jenv, SWIG_RuntimeError, nds_strerror(errno));
		#else
			SWIG_exception(SWIG_RuntimeError, nds_strerror(errno));
		#endif
	%}
%enddef

/** Exception handling for routines that must not set errno. */
%define NDS_SWIG_EXCEPTION_NOERRNO
	%exception %{
		errno = 0;
		$action
		if (errno)
		#if defined(SWIGJAVA)
			SWIG_JavaException(jenv, SWIG_RuntimeError, nds_strerror(errno));
		#else
			SWIG_exception(SWIG_RuntimeError, nds_strerror(errno));
		#endif
	%}
%enddef

/** Trivial exception handling for routines that do not require error checking. */
%define NDS_SWIG_EXCEPTION_NOOP
	%exception;
%enddef

/* Turn on automatic docstrings. */
%feature("autodoc");

/* Disable all default constructors. */
%nodefaultctor;

%newobject _nds_connection::find_channels;
%newobject _nds_connection::fetch;
%newobject _nds_connection::next;

/* Process declarations in the header file. */
%include "nds.h"

/* Add some methods to nds_connection.
 * Sadly, SWIG doesn't create overloads for keyword arguments in Java,
 * so we have to add the overloads by hand. */
struct _nds_connection {
	%extend {

		NDS_SWIG_EXCEPTION_NONNULL
		_nds_connection(const char *host, int port, nds_protocol protocol)
		{
			return nds_open(host, port, protocol);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		_nds_connection(const char *host, int port)
		{
			return nds_open(host, port, NDS_PROTOCOL_TRY);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		char *get_host()
		{
			return nds_get_host($self);
		}

		NDS_SWIG_EXCEPTION_NONNEG
		int get_port()
		{
			return nds_get_port($self);
		}

		NDS_SWIG_EXCEPTION_NONNEG
		nds_protocol get_protocol()
		{
			return nds_get_protocol($self);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_channel ** find_channels(
			size_t *count_channels,
			const char *channel_glob,
			nds_channel_type channel_type_mask,
			nds_data_type data_type_mask,
			float min_smaple_rate,
			float max_sample_rate)
		{
			return nds_find_channels(
				$self,
				count_channels,
				channel_glob,
				channel_type_mask,
				data_type_mask,
				min_smaple_rate,
				max_sample_rate);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_channel ** find_channels(
			size_t *count_channels,
			const char *channel_glob,
			nds_channel_type channel_type_mask,
			nds_data_type data_type_mask)
		{
			return nds_find_channels(
				$self,
				count_channels,
				channel_glob,
				channel_type_mask,
				data_type_mask,
				0.f, 1e12f);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_channel ** find_channels(
			size_t *count_channels,
			const char *channel_glob,
			nds_channel_type channel_type_mask)
		{
			return nds_find_channels(
				$self,
				count_channels,
				channel_glob,
				channel_type_mask,
				(nds_data_type) 0xFF,
				0.f, 1e12f);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_channel ** find_channels(
			size_t *count_channels,
			const char *channel_glob)
		{
			return nds_find_channels(
				$self,
				count_channels,
				channel_glob,
				(nds_channel_type) 0xFF,
				(nds_data_type) 0xFF,
				0.f, 1e12f);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_channel ** find_channels(
			size_t *count_channels)
		{
			return nds_find_channels(
				$self,
				count_channels,
				NULL,
				(nds_channel_type) 0xFF,
				(nds_data_type) 0xFF,
				0.f, 1e12f);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_buffer ** fetch(
			long gps_start,
			long gps_stop,
			const char **channel_names,
			size_t count_channels)
		{
			return nds_fetch(
				$self,
				gps_start,
				gps_stop,
				channel_names,
				count_channels);
		}

		NDS_SWIG_EXCEPTION_NONNEG
		int iterate(
			long gps_start,
			long gps_stop,
			const char **channel_names,
			size_t count_channels
		)
		{
			return nds_iterate(
				$self,
				gps_start,
				gps_stop,
				channel_names,
				count_channels
			);
		}

		NDS_SWIG_EXCEPTION_NONNULL
		c_str __str__()
		{
			char *ret = NULL;
			asprintf(&ret, "<%s:%d (protocol version %d)>",
				nds_get_host($self),
				nds_get_port($self),
				nds_get_protocol($self)
			);
			return ret;
		}

		NDS_SWIG_EXCEPTION_NONNULL
		nds_buffer ** next(
			size_t *count_channels
		)
		{
			return nds_next(
				$self,
				count_channels
			);
		}

		NDS_SWIG_EXCEPTION_NOOP
		~_nds_connection()
		{
			nds_close($self);
		}
	}
};

/* Add string conversion routine for nds_channel. */
%extend nds_channel {

	NDS_SWIG_EXCEPTION_NONNULL
	c_str __str__()
	{
		char *ret = NULL;
		const char *channel_type_str;
		const char *data_type_str;
		switch ($self->channel_type)
		{
			case NDS_CHANNEL_TYPE_ONLINE:
				channel_type_str = "ONLINE";
				break;
			case NDS_CHANNEL_TYPE_RAW:
				channel_type_str = "RAW";
				break;
			case NDS_CHANNEL_TYPE_RDS:
				channel_type_str = "RDS";
				break;
			case NDS_CHANNEL_TYPE_STREND:
				channel_type_str = "STREND";
				break;
			case NDS_CHANNEL_TYPE_MTREND:
				channel_type_str = "MTREND";
				break;
			case NDS_CHANNEL_TYPE_TEST_POINT:
				channel_type_str = "TEST_POINT";
				break;
			case NDS_CHANNEL_TYPE_STATIC:
				channel_type_str = "STATIC";
				break;
			default:
				channel_type_str = NULL;
		}
		switch ($self->data_type)
		{
			case NDS_DATA_TYPE_INT16:
				data_type_str = "INT16";
				break;
			case NDS_DATA_TYPE_INT32:
				data_type_str = "INT32";
				break;
			case NDS_DATA_TYPE_INT64:
				data_type_str = "INT64";
				break;
			case NDS_DATA_TYPE_FLOAT32:
				data_type_str = "FLOAT32";
				break;
			case NDS_DATA_TYPE_FLOAT64:
				data_type_str = "FLOAT64";
				break;
			case NDS_DATA_TYPE_COMPLEX32:
				data_type_str = "COMPLEX32";
				break;
			default:
				data_type_str = NULL;
		}
		asprintf(&ret, "<%s (%gHz, %s, %s)>",
			$self->name,
			(double)$self->sample_rate,
			channel_type_str,
			data_type_str
		);
		return ret;
	}

};

/* Add string conversion routine for nds_buffer. */
%extend nds_buffer {

	NDS_SWIG_EXCEPTION_NONNULL
	c_str __str__()
	{
		char *ret = NULL;
		asprintf(&ret, "<%s (GPS time %ld.%09ld, %ld samples)>",
			$self->channel.name,
			$self->gps_seconds,
			$self->gps_nanoseconds,
			$self->length
		);
		return ret;
	}

};

NDS_SWIG_EXCEPTION_NOOP
